/*------------------------------------------------------------------------------

Copyright 2011 University of Aveiro, Portugal, All Rights Reserved.

These programs are supplied free of charge for research purposes only,
and may not be sold or incorporated into any commercial product. There is
ABSOLUTELY NO WARRANTY of any sort, nor any undertaking that they are
fit for ANY PURPOSE WHATSOEVER. Use them at your own risk. If you do
happen to find a bug, or have modifications to suggest, please report
the same to Armando J. Pinho, ap@ua.pt. The copyright notice above
and this statement of conditions must remain an integral part of each
and every copy made of these files.

------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>

#include "bitio.h"
#include "arith.h"
#include "arith_aux.h"
#include "defs.h"
#include "mem.h"
#include "seqs.h"
#include "model.h"

/*----------------------------------------------------------------------------*/

int main(int argc, char *argv[])

	{
	Symbol s, symbol;
	char verbose, ignEqSize;
	unsigned n, key, maxFailures, keySize;
	unsigned rModelStatus = R_MODEL_OFF;
	int seqSizeDifference;
	FILE *fpOut = NULL;
	RModel *rModel;
	HTable *hTable;
	PModel *pModel;
	Sequence *refSeq, *tarSeq;
	Alphabet *alphabet;

	/* Needed to measure CPU time */
	clock_t tac, start;
	double cpuTimeUsed;

	start = clock();

	verbose = 'n';
	ignEqSize = 'n';
	maxFailures = 0;
	keySize = 11;
	if(argc < 3)
		{
		fprintf(stderr, "Usage: GReEnC [ -o outFile ]\n");
		fprintf(stderr, "              [ -v (print more info) ]\n");
		fprintf(stderr, "              [ -k keySize (def %u) ]\n", keySize);
		fprintf(stderr, "              [ -i (ignore equal size) ]\n");
		fprintf(stderr, "              [ -f maxFailures (def 0) ]\n");
		fprintf(stderr, "              refSeq tarSeq\n");
		exit(1);
		}

	printf("GReEnC version 1.0\n");
	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-k", argv[n]))
			{
			keySize = atoi(argv[n+1]);
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-o", argv[n]) == 0)
			{
			if(!(fpOut = fopen(argv[n+1], "w")))
				{
				fprintf(stderr, "Error: can't open file\n");
				return 1;
				}

			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(!strcmp("-f", argv[n]))
			{
			maxFailures = atoi(argv[n+1]);
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-i", argv[n]) == 0)
			{
			ignEqSize = 'y';
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-v", argv[n]) == 0)
			{
			verbose = 'y';
			break;
			}

	if(!(alphabet = (Alphabet *)Calloc(1, sizeof(Alphabet))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	refSeq = ReadSequence(argv[argc-2]);
	tarSeq = ReadSequence(argv[argc-1]);
	seqSizeDifference = (int)tarSeq->size - refSeq->size;

	printf("The reference sequence has a total of %u bases\n", refSeq->size);
	if(verbose == 'y')
		for(n = 0 ; n < 256 ; n++)
			if(refSeq->usage[n])
				printf("  %c : %u\n", n, refSeq->usage[n]);

	printf("The target sequence has a total of %u bases\n", tarSeq->size);
	if(verbose == 'y')
		for(n = 0 ; n < 256 ; n++)
			if(tarSeq->usage[n])
				printf("  %c : %u\n", n, tarSeq->usage[n]);

	printf("Sequence size difference: %d\n", seqSizeDifference);

	/* Compute the coding alphabet (symbols occurring in the target seq) */
	for(n = 0 ; n < 256 ; n++)
		{
		if(tarSeq->usage[n])
			alphabet->symbols[n] = alphabet->nSymbols++;

		else
			alphabet->symbols[n] = INVALID_SYMBOL;

		}

	printf("The target sequence has %u different symbols\n",alphabet->nSymbols);

	if(!(alphabet->revMap = (Base *)Calloc(alphabet->nSymbols, sizeof(Base))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	if(!(alphabet->probs = (double *)Calloc(alphabet->nSymbols,sizeof(double))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	/*
	 * Compute the probability distribution using the symbol counts of
	 * the target sequence.
	 */
	for(n = 0 ; n < 256 ; n++)
		if(tarSeq->usage[n])
			alphabet->probs[alphabet->symbols[n]] = (tarSeq->usage[n] + 1.0) /
			  (tarSeq->size + alphabet->nSymbols);

	AssignHSymbols(alphabet);
	SetupCaseConv(alphabet);

	rModel = CreateRModel(alphabet->nSymbols);
	hTable = CreateHTable(keySize, H_TABLE_SIZE);
	pModel = CreatePModel(alphabet->nSymbols);

	if(fpOut == NULL)
		fpOut = fopen("/dev/null", "w");

	startoutputtingbits();
	start_encode();

	/* Write several parameters needed by the decoder */
	WriteNBits(tarSeq->size - 1, STORAGE_BITS_SEQUENCE_SIZE, fpOut);
	WriteNBits(alphabet->nSymbols - 1, STORAGE_BITS_N_SYMBOLS, fpOut);
	for(n = 0 ; n < 256 ; n++)
		if(tarSeq->usage[n])
			{
			WriteNBits(n, STORAGE_BITS_SYMBOL, fpOut);
			WriteNBits(tarSeq->usage[n], STORAGE_BITS_USAGE, fpOut);
			}

	WriteNBits(ignEqSize == 'y' ? 1 : 0, STORAGE_BITS_IGN_EQ_SIZE, fpOut);
	WriteNBits(keySize - 1, STORAGE_BITS_KEY_SIZE, fpOut);
	WriteNBits(maxFailures, STORAGE_BITS_MAX_FAILURES, fpOut);

	if(seqSizeDifference != 0 || ignEqSize == 'y')
		{
		/*
		 * Fill the hash table with all kmers and corresponding positions
		 * in the reference sequence.
		 */
		printf("Creating the hash table...\n");
		for(n = 0 ; n < refSeq->size - hTable->keySize ; n++)
			{
			key = GetKey(refSeq->bases + n, keySize, alphabet);
			InsertKeyPos(hTable, key, n);
			if(n % (refSeq->size / 100) == 0)
				fprintf(stderr, "%3d%%\r", (int)(n / (refSeq->size / 100)));

			}

		printf("...done\n");
		}

	/*
	 * Encode
	 */
	printf("Compressing the sequence...\n");
	for(n = 0 ; n < tarSeq->size ; n++)
		{
		symbol = alphabet->symbols[tarSeq->bases[n]];

		/* Check if the rModel should be stopped */
		if(rModelStatus == R_MODEL_ON && (((int)rModel->nTries -
		  (rModel->nHits1 + rModel->nHits2) > maxFailures) ||
		  rModel->pos == refSeq->size))
			rModelStatus = R_MODEL_OFF; /* Stop the model */

		/* Check if the rModel should be restarted */
		if(rModelStatus == R_MODEL_OFF && StartRModel(rModel, hTable,
		  GetKey(tarSeq->bases + n, keySize, alphabet), n,
		  refSeq->size == tarSeq->size && ignEqSize == 'n',
		  abs(seqSizeDifference)))
			rModelStatus = R_MODEL_ON; /* Restart the model */

		pModel->sum = 0;
		if(rModelStatus == R_MODEL_ON)
			{
			ComputeRModelProbs(rModel, refSeq, alphabet);
			for(s = 0 ; s < alphabet->nSymbols ; s++)
				{
				pModel->freqs[s] = 1 + (unsigned)(rModel->probs[s] * MAX_FREQ);
				pModel->sum += pModel->freqs[s];
				}

			}

		else
			{
			for(s = 0 ; s < alphabet->nSymbols ; s++)
				{
				pModel->freqs[s] = 1 + (unsigned)(alphabet->probs[s]* MAX_FREQ);
				pModel->sum += pModel->freqs[s];
				}

			}

		ArithEncodeSymbol(symbol, (int *)(pModel->freqs),
		  (int)pModel->sum, fpOut);
		UpdateRModel(rModel, refSeq, alphabet, symbol);

		if(n % (tarSeq->size / 100) == 0)
			fprintf(stderr, "%3d%%\r", (int)(n / (tarSeq->size / 100)));

		}

	printf("...done\n");

	finish_encode(fpOut);
	doneoutputtingbits(fpOut);
	fclose(fpOut);

	tac = clock();

	printf("Total number of bytes: %llu ( %.5f bpb )\n",
	  _bytes_output, _bytes_output * 8. / tarSeq->size);
	cpuTimeUsed = ((double) (tac - start)) / CLOCKS_PER_SEC;
    printf("Total cpu time used: %g s\n", cpuTimeUsed);
	printf("Total memory allocated by (m/c/re)alloc: %.2f MB\n",
	  TotalMem() / 1024. / 1024.);

	return 0;
	}

