/*------------------------------------------------------------------------------

Copyright 2011 University of Aveiro, Portugal, All Rights Reserved.

These programs are supplied free of charge for research purposes only,
and may not be sold or incorporated into any commercial product. There is
ABSOLUTELY NO WARRANTY of any sort, nor any undertaking that they are
fit for ANY PURPOSE WHATSOEVER. Use them at your own risk. If you do
happen to find a bug, or have modifications to suggest, please report
the same to Armando J. Pinho, ap@ua.pt. The copyright notice above
and this statement of conditions must remain an integral part of each
and every copy made of these files.

------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>

#include "bitio.h"
#include "arith.h"
#include "arith_aux.h"
#include "defs.h"
#include "mem.h"
#include "seqs.h"
#include "model.h"

/*----------------------------------------------------------------------------*/

int main(int argc, char *argv[])

	{
	Symbol s, symbol;
	char verbose, ignEqSize;
	unsigned n, tarSeqSize, key, maxFailures, keySize;
	unsigned rModelStatus = R_MODEL_OFF;
	int seqSizeDifference;
	FILE *fpInp, *fpOut = NULL;
	RModel *rModel;
	HTable *hTable;
	PModel *pModel;
	Sequence *refSeq, *tarSeq;
	Alphabet *alphabet;

	/* Needed to measure CPU time */
	clock_t tac, start;
	double cpuTimeUsed;

	start = clock();

	verbose = 'n';
	if(argc < 3)
		{
		fprintf(stderr, "Usage: GReEnD [ -v (print more info) ]\n");
		fprintf(stderr, "              [ -o outFile ]\n");
		fprintf(stderr, "              refSeq codeFile\n");
		exit(1);
		}

	printf("GReEnD version 1.0\n");
	for(n = 1 ; n < argc ; n++)
		if(strcmp("-v", argv[n]) == 0)
			{
			verbose = 'y';
			break;
			}

	for(n = 1 ; n < argc ; n++)
		if(strcmp("-o", argv[n]) == 0)
			{
			if(!(fpOut = fopen(argv[n+1], "w")))
				{
				fprintf(stderr, "Error: can't open file\n");
				return 1;
				}

			break;
			}

	if(!(fpInp = fopen(argv[argc-1], "r")))
		{
		fprintf(stderr, "Error: unable to open the code file\n");
		return 1;
		}

	if(fpOut == NULL)
		fpOut = fopen("/dev/null", "w");

	if(!(alphabet = (Alphabet *)Calloc(1, sizeof(Alphabet))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	startinputtingbits();
	start_decode(fpInp);

	tarSeqSize = (unsigned)(ReadNBits(STORAGE_BITS_SEQUENCE_SIZE, fpInp)) + 1;

	refSeq = ReadSequence(argv[argc-2]);
	tarSeq = CreateSequence(tarSeqSize);
	seqSizeDifference = (int)tarSeq->size - refSeq->size;

	printf("The reference sequence has a total of %u bases\n", refSeq->size);
	printf("The target sequence has a total of %u bases\n", tarSeq->size);
	printf("Sequence size difference: %d\n", seqSizeDifference);

	alphabet->nSymbols = ReadNBits(STORAGE_BITS_N_SYMBOLS, fpInp) + 1;
	printf("The target sequence has %u different symbols\n",alphabet->nSymbols);

	if(!(alphabet->revMap = (Base *)Calloc(alphabet->nSymbols, sizeof(Base))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	if(!(alphabet->probs = (double *)Calloc(alphabet->nSymbols,sizeof(double))))
		{
		fprintf(stderr, "Error: out of memory\n");
		return 1;
		}

	for(n = 0 ; n < 256 ; n++)
		alphabet->symbols[n] = INVALID_SYMBOL;

	for(n = 0 ; n < alphabet->nSymbols ; n++)
		{
		alphabet->revMap[n] = ReadNBits(STORAGE_BITS_SYMBOL, fpInp);
		alphabet->symbols[alphabet->revMap[n]] = n;
		tarSeq->usage[alphabet->revMap[n]] =
		  ReadNBits(STORAGE_BITS_USAGE, fpInp);
		alphabet->probs[n] = (tarSeq->usage[alphabet->revMap[n]] + 1.0) /
		  (tarSeq->size + alphabet->nSymbols);
		}

	ignEqSize = ReadNBits(STORAGE_BITS_IGN_EQ_SIZE, fpInp) ? 'y' : 'n';
	keySize = ReadNBits(STORAGE_BITS_KEY_SIZE, fpInp) + 1;
	maxFailures = ReadNBits(STORAGE_BITS_MAX_FAILURES, fpInp);

	if(verbose == 'y')
		printf("Max failures: %u\n", maxFailures);

	rModel = CreateRModel(alphabet->nSymbols);
	hTable = CreateHTable(keySize, H_TABLE_SIZE);
	pModel = CreatePModel(alphabet->nSymbols);

	AssignHSymbols(alphabet);
	SetupCaseConv(alphabet);

	if(seqSizeDifference != 0 || ignEqSize == 'y')
		{
		/*
		 * Fill the hash table with all kmers and corresponding positions
		 * in the reference sequence.
		 */
		printf("Creating the hash table...\n");
		for(n = 0 ; n < refSeq->size - hTable->keySize ; n++)
			{
			key = GetKey(refSeq->bases + n, keySize, alphabet);
			InsertKeyPos(hTable, key, n);
			if(n % (refSeq->size / 100) == 0)
				fprintf(stderr, "%3d%%\r", (int)(n / (refSeq->size / 100)));

			}

		printf("...done\n");
		}

	/*
	 * Decode
	 */
    printf("Decompressing the sequence...\n");
	for(n = 0 ; n < tarSeq->size ; n++)
		{
		/* Check if the rModel should be stopped */
		if(rModelStatus == R_MODEL_ON && (((int)rModel->nTries -
		  (rModel->nHits1 + rModel->nHits2) > maxFailures) ||
		  rModel->pos == refSeq->size))
			rModelStatus = R_MODEL_OFF; /* Stop the model */

		/* Check if the rModel should be started */
		if(rModelStatus == R_MODEL_OFF && StartRModel(rModel, hTable,
		  GetKey(tarSeq->bases + n, keySize, alphabet), n,
		  refSeq->size == tarSeq->size && ignEqSize == 'n',
		  abs(seqSizeDifference)))
			rModelStatus = R_MODEL_ON; /* Restart the model */

		pModel->sum = 0;
		if(rModelStatus == R_MODEL_ON)
			{
			ComputeRModelProbs(rModel, refSeq, alphabet);
			for(s = 0 ; s < alphabet->nSymbols ; s++)
				{
				pModel->freqs[s] = 1 + (unsigned)(rModel->probs[s] * MAX_FREQ);
				pModel->sum += pModel->freqs[s];
				}

			}

		else
			{
			for(s = 0 ; s < alphabet->nSymbols ; s++)
				{
				pModel->freqs[s] = 1 + (unsigned)(alphabet->probs[s]* MAX_FREQ);
				pModel->sum += pModel->freqs[s];
				}

			}

		symbol = ArithDecodeSymbol(alphabet->nSymbols, (int *)pModel->freqs,
		  (int)pModel->sum, fpInp);
		tarSeq->bases[n] = alphabet->revMap[symbol];
		fputc(tarSeq->bases[n], fpOut);
		UpdateRModel(rModel, refSeq, alphabet, symbol);

		if(n % (tarSeq->size / 100) == 0)
			fprintf(stderr, "%3d%%\r", (int)(n / (tarSeq->size / 100)));

		}

	printf("...done\n");

	finish_decode();
	doneinputtingbits();
	fclose(fpInp);

	tac = clock();
	cpuTimeUsed = ((double) (tac - start)) / CLOCKS_PER_SEC;
    printf("Got %d bases\n", tarSeq->size);
    printf("Total cpu time used: %g s\n", cpuTimeUsed);

	return 0;
	}

