/*------------------------------------------------------------------------------

Copyright 2011 University of Aveiro, Portugal, All Rights Reserved.

These programs are supplied free of charge for research purposes only,
and may not be sold or incorporated into any commercial product. There is
ABSOLUTELY NO WARRANTY of any sort, nor any undertaking that they are
fit for ANY PURPOSE WHATSOEVER. Use them at your own risk. If you do
happen to find a bug, or have modifications to suggest, please report
the same to Armando J. Pinho, ap@ua.pt. The copyright notice above
and this statement of conditions must remain an integral part of each
and every copy made of these files.

------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "defs.h"
#include "mem.h"
#include "seqs.h"
#include "model.h"

/*----------------------------------------------------------------------------*/

PModel *CreatePModel(unsigned nSymbols)

	{
	PModel *pModel;

	if(!(pModel = (PModel *)Malloc(sizeof(PModel))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!(pModel->freqs = (unsigned *)Malloc(nSymbols * sizeof(unsigned))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	return pModel;
	}

/*----------------------------------------------------------------------------*/

static unsigned SearchClosestPos(HTableEntry *hTableEntry, unsigned curPos)

	{
	int mid = 0, l = 0, u = hTableEntry->nPos - 1;

	if(curPos >= hTableEntry->pos[u])
		return hTableEntry->pos[u];

	if(curPos <= hTableEntry->pos[l])
		return hTableEntry->pos[l];

	while(l <= u)
		{
		mid = (l + u) / 2;
		if(curPos == hTableEntry->pos[mid])
			return curPos;

		else
			if(curPos < hTableEntry->pos[mid])
				{
				u = mid - 1;
				}

			else
				l = mid + 1;

		}

	if(curPos > hTableEntry->pos[mid])
		return curPos-hTableEntry->pos[mid] < hTableEntry->pos[mid+1]-curPos ?
		  hTableEntry->pos[mid] : hTableEntry->pos[mid+1];

	else
		return curPos-hTableEntry->pos[mid-1] < hTableEntry->pos[mid]-curPos ?
		  hTableEntry->pos[mid-1] : hTableEntry->pos[mid];

	}

/*----------------------------------------------------------------------------*/

int StartRModel(RModel *rModel, HTable *hTable, unsigned key, unsigned curPos,
  unsigned useCurPos, unsigned seqSizeAbsDiff)

	{
	unsigned s, closestPos;
	HTableEntry *hTableEntry;

	if(!useCurPos)
		{
		if(!(hTableEntry = GetHTableEntry(hTable, key)))
			return 0;

		closestPos = SearchClosestPos(hTableEntry, curPos);
		rModel->pos = closestPos;
		}

	else
		{
		rModel->pos = curPos;
		}

	rModel->nHits1 = 0;
	rModel->nHits2 = 0;
	rModel->nTries = 0;

	for(s = 0 ; s < rModel->nSymbols ; s++)
		rModel->probs[s] = 0;

	return 1;
	}

/*----------------------------------------------------------------------------*/

RModel *CreateRModel(unsigned nSymbols)

	{
	RModel *rModel;

	if(!(rModel = (RModel *)Calloc(1, sizeof(RModel))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!(rModel->probs = (double *)Calloc(nSymbols, sizeof(double))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	rModel->nSymbols = nSymbols;
	return rModel;
	}

/*----------------------------------------------------------------------------*/

HTable *CreateHTable(unsigned hKeySize, unsigned hSize)

	{
	HTable *hTable;

	if(!(hTable = (HTable *)Calloc(1, sizeof(HTable))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!(hTable->entries = (HTableEntry **)Calloc(hSize,sizeof(HTableEntry *))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	if(!(hTable->entrySize = (unsigned *)Calloc(hSize, sizeof(unsigned))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	hTable->size = hSize;
	hTable->keySize = hKeySize;

	return hTable;
	}

/*------------------------------------------------------------------------------

	"pos" points to the first symbol after the kmer.

------------------------------------------------------------------------------*/

void InsertKeyPos(HTable *hTable, unsigned key, unsigned pos)

	{
	unsigned n;
	unsigned hIdx = key % hTable->size; /* The hash index */

	for(n = 0 ; n < hTable->entrySize[hIdx] ; n++)
		{
		/* If key found */
		if(hTable->entries[hIdx][n].key == key)
			{
			/* Create a new position */
			if(!(hTable->entries[hIdx][n].pos = (unsigned *)
			  Realloc(hTable->entries[hIdx][n].pos,
			  (hTable->entries[hIdx][n].nPos + 1) *
			  sizeof(unsigned), sizeof(unsigned))))
				{
				fprintf(stderr, "Error: in memory allocation\n");
				exit(1);
				}

			hTable->entries[hIdx][n].pos[hTable->entries[hIdx][n].nPos++] = pos;
			return;
			}

		}

	/* If key not found, create a new entry */
	if(!(hTable->entries[hIdx] = (HTableEntry *)Realloc(hTable->entries[hIdx],
	  (hTable->entrySize[hIdx] + 1) * sizeof(HTableEntry),sizeof(HTableEntry))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	/* Create a new position */
	if(!(hTable->entries[hIdx][hTable->entrySize[hIdx]].pos =
	  (unsigned *)Malloc(sizeof(unsigned))))
		{
		fprintf(stderr, "Error: in memory allocation\n");
		exit(1);
		}

	hTable->entries[hIdx][hTable->entrySize[hIdx]].key = key;
	hTable->entries[hIdx][hTable->entrySize[hIdx]].pos[0] = pos;
	hTable->entries[hIdx][hTable->entrySize[hIdx]].nPos = 1;

	hTable->entrySize[hIdx]++;
	}

/*----------------------------------------------------------------------------*/

HTableEntry *GetHTableEntry(HTable *hTable, unsigned key)

	{
	unsigned n;
	unsigned hIdx = key % hTable->size; /* The hash index */

	for(n = 0 ; n < hTable->entrySize[hIdx] ; n++)
		if(hTable->entries[hIdx][n].key == key)
			return &hTable->entries[hIdx][n];

	return NULL;
	}

/*----------------------------------------------------------------------------*/

void ComputeRModelProbs(RModel *rModel, Sequence *refSeq, Alphabet *alphabet)

	{
	Base refBase = refSeq->bases[rModel->pos];
	Symbol s, symbol1, symbol2;
	double correctionFactor;

	symbol1 = alphabet->symbols[refBase];
	symbol2 = alphabet->symbols[alphabet->caseConv[refBase]];

	if(symbol1 == INVALID_SYMBOL)
		{
		if(symbol2 == INVALID_SYMBOL)
			{
			for(s = 0 ; s < alphabet->nSymbols ; s++)
				rModel->probs[s] = alphabet->probs[s];

			}

		else
			{
			rModel->probs[symbol2] = (rModel->nHits2 + 1.0) /
			  (rModel->nTries + 2.0);
			correctionFactor = (1 - rModel->probs[symbol2]) /
			  (1 - alphabet->probs[symbol2]);
			for(s = 0 ; s < alphabet->nSymbols ; s++)
				if(s != symbol2)
					rModel->probs[s] = alphabet->probs[s] * correctionFactor;

			}

		}

	else
		{
		if(symbol2 == INVALID_SYMBOL)
			{
			rModel->probs[symbol1] = (rModel->nHits1 + 1.0) /
			  (rModel->nTries + 2.0);

			correctionFactor = (1 - rModel->probs[symbol1]) /
			  (1 - alphabet->probs[symbol1]);

			for(s = 0 ; s < alphabet->nSymbols ; s++)
				if(s != symbol1)
					rModel->probs[s] = alphabet->probs[s] * correctionFactor;

			}

		else
			{
			rModel->probs[symbol1] = (rModel->nHits1 + 1.0) /
			  (rModel->nTries + 3.0);
			rModel->probs[symbol2] = (rModel->nHits2 + 1.0) /
			  (rModel->nTries + 3.0);

			correctionFactor = (1 - rModel->probs[symbol1] -
			  rModel->probs[symbol2]) / (1 - alphabet->probs[symbol1] -
			  alphabet->probs[symbol2]);

			for(s = 0 ; s < alphabet->nSymbols ; s++)
				if(s != symbol1 && s != symbol2)
					rModel->probs[s] = alphabet->probs[s] * correctionFactor;

			}

		}

	}

/*----------------------------------------------------------------------------*/

void UpdateRModel(RModel *rModel, Sequence *refSeq, Alphabet *alphabet,
  Symbol symbol)

	{
	Base refBase = refSeq->bases[rModel->pos++];
	Symbol refSymbol = alphabet->symbols[refBase];

	if(refSymbol == symbol)
		rModel->nHits1++;

	else
		if(alphabet->symbols[alphabet->caseConv[refBase]] == symbol)
			rModel->nHits2++;

	rModel->nTries++;
	}

